/* csol
 * Copyright (c) 2017 Niels Sonnich Poulsen (http://nielssp.dk)
 * Licensed under the MIT license.
 * See the LICENSE file or http://opensource.org/licenses/MIT for more information.
 */

#define _GNU_SOURCE
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#include "theme.h"
#include "util.h"
#include "rc.h"

struct dir_list {
  char *dir;
  struct dir_list *next;
};

ThemeList *first_theme = NULL;
ThemeList *last_theme = NULL;

struct dir_list *theme_dirs = NULL;

Theme *new_theme() {
  Theme *theme = malloc(sizeof(Theme));
  theme->name = NULL;
  theme->title = NULL;
  theme->heart = NULL;
  theme->spade = NULL;
  theme->diamond = NULL;
  theme->club = NULL;
  theme->width = 6;
  theme->height = 4;
  theme->x_spacing = 2;
  theme->y_spacing = 1;
  theme->x_margin = 2;
  theme->y_margin = 1;
  theme->colors = NULL;
  theme->background = (ColorPair){.fg = 7, .bg = 0};
  theme->empty_layout = init_layout();
  theme->back_layout = init_layout();
  theme->red_layout = init_layout();
  theme->black_layout = init_layout();
  return theme;
}

Layout init_layout() {
  return (Layout){
    .color = (ColorPair){.fg = 7, .bg = 0},
    .top = NULL,
    .middle = NULL,
    .bottom = NULL,
    .text_fields = NULL
  };
}

void define_color(Theme *theme, short index, short red, short green, short blue) {
  Color *color = malloc(sizeof(Color));
  color->next = theme->colors;
  color->index = index;
  color->red = red;
  color->green = green;
  color->blue = blue;
  theme->colors = color;
}

void register_theme(Theme *theme) {
  if (theme->name) {
    ThemeList *next = malloc(sizeof(struct theme_list));
    next->theme = theme;
    next->next = NULL;
    if (last_theme) {
      last_theme->next = next;
      last_theme = next;
    } else {
      first_theme = last_theme = next;
    }
  }
}

void register_theme_dir(const char *cwd, const char *dir) {
  struct dir_list *theme_dir = malloc(sizeof(struct dir_list));
  theme_dir->dir = combine_paths(cwd, dir);
  theme_dir->next = theme_dirs;
  theme_dirs = theme_dir;
}

void load_theme_dirs() {
  struct dir_list *current = theme_dirs;
  while (current) {
    struct dirent **files;
    int n = scandir(current->dir, &files, NULL, alphasort);
    if (n >= 0) {
      for (int i = 0; i < n; i++) {
        char *theme_path = combine_paths(current->dir, files[i]->d_name);
        execute_file(theme_path);
        free(theme_path);
        free(files[i]);
      }
      free(files);
    }
    struct dir_list *next = current->next;
    free(current->dir);
    free(current);
    current = next;
  }
  theme_dirs = NULL;
}

ThemeList *list_themes() {
  return first_theme;
}

Theme *get_theme_in_list(const char *name) {
  for (ThemeList *themes = list_themes(); themes; themes = themes->next) {
    if (strcmp(themes->theme->name, name) == 0) {
      return themes->theme;
    }
  }
  return NULL;
}

Theme *get_theme(const char *name) {
  Theme *theme = get_theme_in_list(name);
  if (theme) {
    return theme;
  }
  for (struct dir_list *theme_dir = theme_dirs; theme_dir; theme_dir = theme_dir->next) {
    char *theme_path = combine_paths(theme_dir->dir, name);
    if (file_exists(theme_path)) {
      execute_file(theme_path);
      theme = get_theme_in_list(name);
      if (theme) {
        free(theme_path);
        return theme;
      }
    }
    free(theme_path);
  }
  return NULL;
}

char *card_suit(Card *card, Theme *theme) {
#ifdef ZIPIT_Z2
  // Substitute for utf-8 suit strings.
  // In case we ssh into the zipit and use large theme for a bigger terminal window.
  switch (card->suit) {
    case HEART:
      if (strlen(theme->heart) > 1)
        return "H";
      return theme->heart;
    case SPADE:
      if (strlen(theme->spade) > 1)
        return "S";
      return theme->spade;
    case DIAMOND:
      if (strlen(theme->diamond) > 1)
        return "D";
      return theme->diamond;
    case CLUB:
      if (strlen(theme->club) > 1)
        return "C";
      return theme->club;
  }
#else
  switch (card->suit) {
    case HEART:
      return theme->heart;
    case SPADE:
      return theme->spade;
    case DIAMOND:
      return theme->diamond;
    case CLUB:
      return theme->club;
  }
#endif
  return "";
}
